"""Formatter subclasses for different formats."""

from .fmt_csv import CSVFormatter
from .fmt_json import JSONFormatter
from .fmt_table import TableFormatter


_FORMATTERS = {
    formatter.fmt: formatter
    for formatter in (CSVFormatter, JSONFormatter, TableFormatter)}


def get_formats():
    """Return a sorted list of available formatters names."""
    return sorted(_FORMATTERS)


def get_formatter(name):
    """Return the formatter class for the specified format."""
    return _FORMATTERS[name]



"""CSV formatter."""

from csv import writer

from ..formatter import Formatter


class CSVFormatter(Formatter):
    """Format fields in CSV.

    Config parameters:

    - tabs: use tabs as separators
    """

    fmt = 'csv'

    config = {'tabs': False}

    def __init__(self, stream, fields, **kwargs):
        super().__init__(stream, fields, **kwargs)
        dialect = 'excel-tab' if self._config['tabs'] else 'excel'
        self._writer = writer(self._stream, dialect=dialect)

    def _format_header(self):
        self._writer.writerow(self.fields)

    def _format_process(self, process):
        self._writer.writerow(self._fields_values(process))



"""JSON formatter."""

import json

from ..formatter import Formatter


class JSONFormatter(Formatter):
    """Format fields in JSON.

    Config parameters:

      - indent: the amount of spaces for indentation (no indentation if None)
    """

    fmt = 'json'

    config = {'indent': None}

    def __init__(self, stream, fields, **kwargs):
        super().__init__(stream, fields, **kwargs)
        self._data = {'fields': self.fields, 'processes': []}

    def _format_process(self, process):
        self._data['processes'].append(
            dict(zip(self.fields, self._fields_values(process))))

    def _dump(self):
        json.dump(self._data, self._stream, indent=self._config['indent'])



from os import path

from toolrack.collect import Collection

from ...testing import TestCase
from ...tracing.tracer import (
    UnsupportedTracer,
    Tracing,
    Tracer)
from ...tracing.types import TracerType


class TracingTests(TestCase):

    def setUp(self):
        super().setUp()
        self.tracing = Tracing(path=self.tempdir.path)
        self.tempdir.mkdir(path='instances')

    def test_tracers_empty(self):
        """If no tracer is defined, an empty list is returned."""
        self.assertEqual(self.tracing.tracers, [])

    def test_tracers_exists(self):
        """A Tracer object is returned for each tracing directory."""
        self.tempdir.mkdir(path='instances/tracer-1')
        self.tempdir.mkdir(path='instances/tracer-2')
        for tracer in self.tracing.tracers:
            self.assertIsInstance(tracer, Tracer)
        self.assertEqual(
            [tracer.name for tracer in self.tracing.tracers],
            ['tracer-1', 'tracer-2'])

    def test_get_tracer_existing(self):
        """A Tracer for a specified tracing instance is returned."""
        self.tempdir.mkdir(path='instances/tracer')
        tracer = self.tracing.get_tracer('tracer')
        self.assertEqual(tracer.name, 'tracer')

    def test_get_tracer_create(self):
        """If the requested tracing instance doesn't exist, it's created."""
        self.tracing.get_tracer('tracer')
        instance_path = path.join(self.tempdir.path, 'instances/tracer')
        self.assertTrue(path.exists(instance_path))

    def test_remove_tracer(self):
        """A Tracer can be removed."""
        self.tempdir.mkdir(path='instances/tracer')
        instance_path = path.join(self.tempdir.path, 'instances/tracer')
        self.tracing.remove_tracer('tracer')
        self.assertFalse(path.exists(instance_path))


class TracerTests(TestCase):

    def setUp(self):
        super().setUp()
        self.tracer = Tracer(path=self.tempdir.path)

    def test_name(self):
        """The Tracer name is the name of the tracer directory."""
        name = path.basename(self.tempdir.path)
        self.assertEqual(self.tracer.name, name)

    def test_type(self):
        """The Tracer type can be returned."""
        self.tempdir.mkfile(path='current_tracer', content='nop')
        self.assertEqual(self.tracer.type, 'nop')

    def test_set_type(self):
        """The Tracer type can be set."""
        self.tempdir.mkfile(path='current_tracer', content='func')
        self.tracer.set_type('nop')
        self.assertEqual(self.tracer.type, 'nop')

    def test_set_type_unsupported(self):
        """If the Tracer type is unsupported, an error is raised."""
        self.assertRaises(UnsupportedTracer, self.tracer.set_type, 'unknown')

    def test_enabled_true(self):
        """The Tracer is enabled if the corresponding flag is set."""
        self.tempdir.mkfile(path='tracing_on', content='1')
        self.assertTrue(self.tracer.enabled)

    def test_enabled_false(self):
        """The Tracer is enabled if the corresponding flag is not set."""
        self.tempdir.mkfile(path='tracing_on', content='0')
        self.assertFalse(self.tracer.enabled)

    def test_toggle(self):
        """The Tracer can be enabled and disabled."""
        self.tempdir.mkfile(path='tracing_on', content='0')
        self.tracer.toggle(True)
        self.assertTrue(self.tracer.enabled)
        self.tracer.toggle(False)
        self.assertFalse(self.tracer.enabled)

    def test_options(self):
        """Tracer options are returned, with their status."""
        self.tempdir.mkfile(path='trace_options', content='noraw\nhex')
        self.assertEqual({'raw': False, 'hex': True}, self.tracer.options)

    def test_set_option(self):
        """Tracer options can be set."""
        self.tempdir.mkfile(path='trace_options', content='nohex')
        self.assertEqual({'hex': False}, self.tracer.options)
        self.tracer.set_option('hex', True)
        self.assertEqual({'hex': True}, self.tracer.options)
        self.tracer.set_option('hex', False)
        self.assertEqual({'hex': False}, self.tracer.options)

    def test_trace_content(self):
        """The trace file content can be returned."""
        self.tempdir.mkfile(path='trace', content='some trace content')
        self.assertEqual('some trace content', self.tracer.trace())

    def test_trace_pipe(self):
        """The trace_pipe file cam be returned and read."""
        self.tempdir.mkfile(path='trace_pipe', content='some trace content')
        with self.tracer.trace_pipe() as pipe:
            self.assertEqual('some trace content', pipe.read())

    def test_attribute_access(self):
        """Attribute specific to the tracer type can be accessed."""

        class SampleTracer(TracerType):

            name = 'sample'

            foo = 'a sample attribute'

        self.tracer._tracer_types = Collection('TracerType', 'name')
        self.tracer._tracer_types.add(SampleTracer)

        self.tempdir.mkfile(path='current_tracer')
        self.tracer.set_type('sample')
        self.assertEqual(self.tracer.foo, 'a sample attribute')



"""Write field values for a collection of processes in a specific format."""


class Formatter:
    """Format a :class:`Process` :class:`Collection`.

    Parameters:
      stream: a file-like stream to write the formatted output to.
      fields: a list of Process attributes to print.
      Kwargs: filter configuration parameters.

    """

    # Name of the format produced by the formatter.
    fmt = ''

    # Configuration parameters with defaults.
    config = {}

    def __init__(self, stream, fields, **kwargs):
        self._stream = stream
        self.fields = fields

        unknown_keys = set(kwargs).difference(self.config)
        if unknown_keys:
            raise TypeError(
                'Unknown config parameters: {}'.format(
                    ', '.join(sorted(unknown_keys))))
        self._config = self.config.copy()
        self._config.update(kwargs)

    def _fields_values(self, process):
        """Return a list of fields values for a :class:`Process`."""
        return [process.get(field) for field in self.fields]

    def format(self, collection):
        """Write the formatted output of the :class:`Collection`."""
        self._format_header()
        for process in collection:
            self._format_process(process)
        self._format_footer()
        self._dump()

    def _format_header(self):
        """Format the header.

        Subclasses can implement this.

        """
        pass

    def _format_process(self, process):
        """Format data for a :class:`Process`.

        Subclasses can implement this.

        """
        pass

    def _format_footer(self):
        """Format footer.

        Subclasses can implement this.

        """
        pass

    def _dump(self):
        """Dump processed data to the stream.

        Subclasses can implement this.

        """
        pass

    def _write(self, data):
        if data is not None:
            self._stream.write(data)



from io import StringIO

from ...testing import TestCase

from ..formatter import Formatter
from ..collection import (
    Collection,
    Collector)
from ..process import Process


class SampleFormatter(Formatter):

    config = {'option': 10}
    endl = '\n'

    def _format_header(self):
        self._write('header\n')

    def _format_footer(self):
        self._write('footer\n')

    def _format_process(self, process):
        self._write('process {} {}\n'.format(process.pid, process.cmd))

    def _dump(self):
        self._write('dump\n')


class FormatterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.stream = StringIO()

    def test_format_no_op_default(self):
        """By default, format methods don't output anything."""
        self.make_process_file(10, 'cmdline', content='cmd1')
        self.make_process_file(20, 'cmdline', content='cmd2')
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        collection = Collection(collector=collector)
        formatter = Formatter(self.stream, ['pid'])
        formatter.format(collection)
        self.assertEqual(self.stream.getvalue(), '')

    def test_format(self):
        """Formatter.format outputs process info with header and footer."""
        self.make_process_file(10, 'cmdline', content='cmd1')
        self.make_process_file(20, 'cmdline', content='cmd2')
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        collection = Collection(collector=collector)
        formatter = SampleFormatter(self.stream, ['pid', 'cmd'])
        formatter.format(collection)
        self.assertEqual(
            self.stream.getvalue(),
            'header\n'
            'process 10 cmd1\n'
            'process 20 cmd2\n'
            'footer\n'
            'dump\n')

    def test_fields_values(self):
        """Formatter._fields_values returns a list with Process values."""
        self.make_process_file(10, 'cmdline', '/bin/foo')
        formatter = SampleFormatter(self.stream, ['pid', 'cmd'])
        process = Process(10, proc_dir='{}/10'.format(self.tempdir.path))
        process.collect_stats()
        self.assertEqual(
            formatter._fields_values(process), [10, '/bin/foo'])

    def test_config_default_value(self):
        """Formatter can have config options with default values."""
        formatter = SampleFormatter(self.stream, ['pid', 'cmdline'])
        self.assertEqual(formatter._config, {'option': 10})

    def test_config_set_value(self):
        """It's possible to specify config options to Formatter."""
        formatter = SampleFormatter(self.stream, ['pid', 'cmdline'], option=30)
        self.assertEqual(formatter._config, {'option': 30})

    def test_unknown_option(self):
        """If an unknown option is specified, an error is raised."""
        self.assertRaises(
            TypeError, SampleFormatter, self.stream, ['pid', 'cmdline'],
            unknown_option=30)



from ...testing import TestCase
from ..process import Process
from ..filter import (
    CommandNameFilter,
    CommandLineFilter)


class CommandNameFilterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.pid = 10
        self.process = Process(
            self.pid, proc_dir='{}/{}'.format(self.tempdir.path, self.pid))

    def test_filter_matching_command(self):
        """CommandNameFilter matches a process."""
        proc_filter = CommandNameFilter('foo')
        self.make_process_file(
            self.pid, 'cmdline', content='foo\x00bar\x00')
        self.process.collect_stats()
        self.assertTrue(proc_filter(self.process))

    def test_filter_only_exact(self):
        """CommandNameFilter only matches exact command names."""
        proc_filter = CommandNameFilter('foo')
        self.make_process_file(
            self.pid, 'cmdline', content='/bin/foo\x00')
        self.process.collect_stats()
        self.assertFalse(proc_filter(self.process))

    def test_filter_matching_comm(self):
        """CommandNameFilter matches a process by comm value."""
        proc_filter = CommandNameFilter('foo')
        self.make_process_file(self.pid, 'comm', content='foo')
        self.process.collect_stats()
        self.assertTrue(proc_filter(self.process))


class CommandLineFilterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.pid = 10
        self.process = Process(
            self.pid, proc_dir='{}/{}'.format(self.tempdir.path, self.pid))

    def test_filter_matching_process(self):
        """CommandLineFilter matches a process."""
        proc_filter = CommandLineFilter('foo')
        self.make_process_file(
            self.pid, 'cmdline', content='/bin/foo\x00bar\x00')
        self.process.collect_stats()
        self.assertTrue(proc_filter(self.process))

    def test_filter_non_matching_process(self):
        """CommandLineFilter doesn't match a process."""
        proc_filter = CommandLineFilter('foo')
        self.make_process_file(
            self.pid, 'cmdline', content='/bin/bar\x00foo\x00')
        self.process.collect_stats()
        self.assertFalse(proc_filter(self.process))

    def test_include_args(self):
        """Arguments are included in match if include_args is True."""
        proc_filter = CommandLineFilter('foo', include_args=True)
        self.make_process_file(
            self.pid, 'cmdline', content='/bin/bar\x00foo\x00')
        self.process.collect_stats()
        self.assertTrue(proc_filter(self.process))



from io import StringIO

from ....testing import TestCase
from .. import CSVFormatter
from ...collection import (
    Collector,
    Collection)


class CSVFormatterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.stream = StringIO()
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        self.collection = Collection(collector=collector)
        self.make_process_file(10, 'cmdline', content='/bin/foo')
        self.make_process_file(20, 'cmdline', content='/bin/bar')

    def test_format(self):
        """CSVFormatter outputs CSV file with processes data."""
        formatter = CSVFormatter(self.stream, ['pid', 'cmd'])
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            'pid,cmd\r\n'
            '10,/bin/foo\r\n'
            '20,/bin/bar\r\n')

    def test_format_tabs(self):
        """CSVFormatter can use tabs as field separators."""
        formatter = CSVFormatter(self.stream, ['pid', 'cmd'], tabs=True)
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            'pid\tcmd\r\n'
            '10\t/bin/foo\r\n'
            '20\t/bin/bar\r\n')



from textwrap import dedent

from ....testing import TestCase
from ..system import (
    ProcStat,
    ProcUptime,
    ProcLoadavg,
    ProcVmstat,
    ProcDiskstats,
    ProcMeminfo,
    ProcCgroups)


class ProcStatTests(TestCase):

    def test_cpu_fields(self):
        """Time counters are reported as percentage per-CPU."""
        path = self.tempdir.mkfile(
            content=(
                'cpu0 2.0 0.0 3.0 5.0 7.0 3.0 10.0 20.0 30.0 20.0\n'
                'cpu1 10.0 7.0 5.0 3.0 0.0 3.0 2.0 30.0 20.0 20.0\n'))
        stat_file = ProcStat(path)
        self.assertEqual(
            stat_file.parse(),
            {'cpu0': {
                'user': 0.02, 'nice': 0.0, 'system': 0.03, 'idle': 0.05,
                'iowait': 0.07, 'irq': 0.03, 'softirq': 0.1, 'steal': 0.2,
                'guest': 0.3, 'guest-nice': 0.2},
             'cpu1': {
                'user': 0.1, 'nice': 0.07, 'system': 0.05, 'idle': 0.03,
                'iowait': 0.00, 'irq': 0.03, 'softirq': 0.02, 'steal': 0.3,
                'guest': 0.2, 'guest-nice': 0.2}})

    def test_not_all_cpu_fields(self):
        """Some CPU fields may be missing."""
        path = self.tempdir.mkfile(
            content='cpu0 20.0 10.0 30.0 20.0 7.0 3.0 10.0\n')
        stat_file = ProcStat(path)
        self.assertEqual(
            stat_file.parse(),
            {'cpu0': {
                'user': 0.2, 'nice': 0.1, 'system': 0.3, 'idle': 0.2,
                'iowait': 0.07, 'irq': 0.03, 'softirq': 0.1}})

    def test_ignore_non_cpu_fields(self):
        """Extra fields after CPU ones are ignored."""
        path = self.tempdir.mkfile(
            content=(
                'cpu0 20.0 10.0 30.0 20.0 7.0 3.0 10.0\nanother row\n'))
        stat_file = ProcStat(path)
        self.assertEqual(
            stat_file.parse(),
            {'cpu0': {
                'user': 0.2, 'nice': 0.1, 'system': 0.3, 'idle': 0.2,
                'iowait': 0.07, 'irq': 0.03, 'softirq': 0.1}})


class ProcUptimeTests(TestCase):

    def test_fields(self):
        """Uptime and idle times are reported."""
        path = self.tempdir.mkfile(content='67569.47 106913.77')
        uptime_file = ProcUptime(path)
        self.assertEqual(
            uptime_file.parse(), {'uptime': 67569.47, 'idle': 106913.77})


class ProcLoadavgTests(TestCase):

    def test_fields(self):
        """Load average over 1, 5, and 15 minutes is reported."""
        path = self.tempdir.mkfile(content='0.40 0.30 0.20')
        loadavg_file = ProcLoadavg(path)
        self.assertEqual(
            loadavg_file.parse(),
            {'load1': 0.40, 'load5': 0.30, 'load15': 0.20})


class ProcVmstatTests(TestCase):

    def test_fields(self):
        """Fields and values from the /proc/vmstat file are reported."""
        path = self.tempdir.mkfile(content='foo 123\nbar 456')
        vmstat_file = ProcVmstat(path)
        self.assertEqual(vmstat_file.parse(), {'foo': 123, 'bar': 456})


class ProcDiskstatsTests(TestCase):

    def test_fields(self):
        """Fields for each device/partition are reported."""
        path = self.tempdir.mkfile(
            content=dedent(
                '''\
                8 0 sda 10 20 30 40 50 60 70 80 90 100 110
                8 1 sda1 1 2 3 4 5 6 7 8 9 10 11
                '''))
        diskstats_file = ProcDiskstats(path)
        self.assertEqual(
            diskstats_file.parse(),
            {'sda': {
                'read': 10, 'read-merged': 20, 'read-sect': 30, 'read-ms': 40,
                'write': 50, 'write-merged': 60, 'write-sect': 70,
                'write-ms': 80, 'io-curr': 90, 'io-ms': 100,
                'io-ms-weighted': 110},
             'sda1': {
                'read': 1, 'read-merged': 2, 'read-sect': 3, 'read-ms': 4,
                'write': 5, 'write-merged': 6, 'write-sect': 7, 'write-ms': 8,
                'io-curr': 9, 'io-ms': 10, 'io-ms-weighted': 11}})


class ProcMeminfoTests(TestCase):

    def test_fields(self):
        """Each line in the file /proc/meminfo file is reported."""
        path = self.tempdir.mkfile(
            content=dedent(
                '''\
                MemTotal:       1000 kB
                MemFree:         200 kB
                HugePages_Total:   0
                '''))
        meminfo_file = ProcMeminfo(path)
        self.assertEqual(
            meminfo_file.parse(),
            {'MemTotal': 1000, 'MemFree': 200, 'HugePages_Total': 0})


class ProcCgroupsTests(TestCase):

    def test_fields(self):
        """Each line in the file /proc/cgroups file is reported."""
        path = self.tempdir.mkfile(
            content=dedent(
                '''\
                #subsys_name    hierarchy       num_cgroups     enabled
                cpuset  10      200     1
                cpu     3       100     0
                cpuacct 3       300     1
                '''))
        cgroups_file = ProcCgroups(path)
        self.assertEqual(
            cgroups_file.parse(),
            {'cpuset': {
                'hierarchy-id': 10, 'num-cgroups': 200, 'enabled': True},
             'cpu': {
                 'hierarchy-id': 3, 'num-cgroups': 100, 'enabled': False},
             'cpuacct': {
                'hierarchy-id': 3, 'num-cgroups': 300,  'enabled': True}})



"""ps-like script to dump processes information."""

import os
import sys
from itertools import repeat
from time import sleep
from argparse import (
    ArgumentParser,
    ArgumentTypeError)

from toolrack.script import Script

from ..process.collection import (
    Collection,
    Collector)
from ..process.filter import CommandLineFilter
from ..process.formatters import (
    get_formats,
    get_formatter)


class ProcsScript(Script):
    """ps-like utility.

    It supports commandline-based filtering and output in different formats.

    """

    def get_parser(self):
        parser = ArgumentParser(
            description='Dump info about running processes.')

        def pids(pid_list):
            """Comma-separated list of PIDs."""
            try:
                return [int(pid) for pid in pid_list.split(',')]
            except:
                raise ArgumentTypeError('Must specify a list of PIDs')

        parser.add_argument(
            '--available-stats', help='Print a list of available stats.',
            action='store_true')
        parser.add_argument(
            '--fields', '-f',
            help='comma-separated list of fields to display',
            default='pid,stat.state,comm')
        parser.add_argument(
            '--regexp', '-r', help='regexp to filter by process name')
        parser.add_argument(
            '--cmdline-regexp', '-R',
            help='regexp to filter by full command line')
        parser.add_argument(
            '--pids', '-p', help='list specific PIDs', type=pids)
        parser.add_argument(
            '--format', '-F', help='output format', choices=get_formats(),
            default='table')
        parser.add_argument(
            '--interval', '-i',
            help='sample interval in seconds (default %(default)s)', type=int,
            default=5)
        parser.add_argument(
            '--count', '-c',
            help='number of samples to collect (default unlimited)', type=int,
            default=0)
        return parser

    def main(self, args):
        if args.available_stats:
            self._print_available_stats()
            self.exit()

        fields = [field.strip() for field in args.fields.split(',')]

        collector = Collector(pids=args.pids)
        collection = Collection(collector=collector)
        if args.regexp:
            collection.add_filter(CommandLineFilter(args.regexp))
        if args.cmdline_regexp:
            collection.add_filter(
                CommandLineFilter(args.cmdline_regexp, include_args=True))
        formatter_class = get_formatter(args.format)
        formatter = formatter_class(sys.stdout, fields)

        count_iter = range(args.count) if args.count else repeat(None)
        for n in count_iter:
            formatter.format(collection)
            if n != args.count - 1:
                # don't sleep after last iteration
                sleep(args.interval)

    def _print_available_stats(self):
        collector = Collector(pids=[os.getpid()])
        collection = Collection(collector=collector)
        [process] = collection
        for stat in process.available_stats():
            print(stat)


script = ProcsScript()



"""Filter classes for process Collection."""

import re


class CommandNameFilter:
    """Filter processes based on the command name.

    Parameters:
      name: the name of the command.

    """

    def __init__(self, name):
        self._name = name

    def __call__(self, process):
        cmdline = process.get('cmdline')
        if cmdline:
            return cmdline[0] == self._name

        return process.get('comm') == self._name


class CommandLineFilter:
    """Filter Processes based on the command line.

    Parameters:
      regexp: a regexp to match the commandline.
      include_args: whether include args in match.

    """

    def __init__(self, regexp, include_args=False):
        self._re = re.compile(regexp)
        self._include_args = include_args

    def __call__(self, process):
        cmd = process.cmd
        if not self._include_args:
            cmd = cmd.split()[0]
        return bool(self._re.findall(cmd))



from unittest import TestCase

from ...files.sys import TracingDirectory
from ...tracing.types import (
    TracerType,
    TRACER_TYPES,
    NopTracer)


class TracerTypeTests(TestCase):

    def test_path(self):
        """A TracerType has a path and a TracingDirectory pointing to it."""
        tracer_type = TracerType('/foo/bar')
        self.assertEqual(tracer_type.path, '/foo/bar')
        self.assertIsInstance(tracer_type._dir, TracingDirectory)
        self.assertEqual(tracer_type._dir.path, '/foo/bar')


class TracerTypesTyests(TestCase):

    def test_tracer_types_registry(self):
        """TRACER_TYPES is a registry of available TracerTypes."""
        self.assertIs(TRACER_TYPES.get('nop'), NopTracer)



"""Table formatter."""

from prettytable import PrettyTable

from ..formatter import Formatter


class TableFormatter(Formatter):
    """Format fields as a table.

    Config parameters:

    - borders: whether to print table borders
    """

    fmt = 'table'

    config = {'borders': False}

    def __init__(self, stream, fields, **kwargs):
        super().__init__(stream, fields, **kwargs)
        self._table = None

    def _format_header(self):
        self._table = PrettyTable()
        for field in self.fields:
            self._table.add_column(field, [], align='l')

    def _format_process(self, process):
        self._table.add_row(self._fields_values(process))

    def _dump(self):
        content = self._table.get_string(border=self._config['borders'])
        self._write(content + '\n')
        self._table = None



"""Linux kernel tracer types."""

from toolrack.collect import Collection

from ..files.sys import TracingDirectory


# Available tracer types
TRACER_TYPES = Collection('TracerType', 'name')


class TracerType:
    """Base class for tracer types."""

    name = None

    def __init__(self, path):
        self.path = path
        self._dir = TracingDirectory(path)


@TRACER_TYPES.add
class NopTracer(TracerType):
    """No-op tracer."""

    name = 'nop'



from ...testing import TestCase
from ..types import (
    OptionsFile,
    SelectableOptionsFile,
    TogglableOptionsFile,
    ValueFile,
    ToggleFile)


class OptionsFileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.options_file = OptionsFile(self.tempdir.join(self.filename))

    def test_options_single_line(self):
        """Option values can be listed from single-line files."""
        self.tempdir.mkfile(path=self.filename, content='foo bar baz')
        self.assertEqual(self.options_file.options, ['foo', 'bar', 'baz'])

    def test_options_multiple_lines(self):
        """Option values can be listed from multi-lines files."""
        self.tempdir.mkfile(path=self.filename, content='foo\nbar\nbaz')
        self.assertEqual(self.options_file.options, ['foo', 'bar', 'baz'])

    def test_options_selected(self):
        """The selected option is included."""
        self.tempdir.mkfile(path=self.filename, content='foo [bar] baz')
        self.assertEqual(self.options_file.options, ['foo', 'bar', 'baz'])


class SelectableOptionsFileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.select_file = SelectableOptionsFile(
            self.tempdir.join(self.filename))

    def test_selected(self):
        """The selected option can be returned."""
        self.tempdir.mkfile(path=self.filename, content='foo [bar] baz')
        self.assertEqual(self.select_file.selected, 'bar')

    def test_selected_none(self):
        """None is returned if no option is selected."""
        self.tempdir.mkfile(path=self.filename, content='foo bar baz')
        self.assertIsNone(self.select_file.selected)

    def test_select(self):
        """The specified option can be selected."""
        self.tempdir.mkfile(path=self.filename, content='foo [bar] baz')
        self.select_file.select('baz')
        # The value is written to file. For /proc and /sys files this will
        # result in passing the value to the kernel.
        self.assertEqual(
            self.readfile(self.tempdir.join(self.filename)), 'baz')

    def test_select_invalid(self):
        """If an invalid value is specified, a ValueError is raised."""
        self.tempdir.mkfile(path=self.filename, content='foo [bar] baz')
        self.assertRaises(ValueError, self.select_file.select, 'unknown')


class TogglableOptionsFileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.toggle_options_file = TogglableOptionsFile(
            self.tempdir.join(self.filename))

    def test_options(self):
        """Options are returned as a dict with current values."""
        self.tempdir.mkfile(path=self.filename, content='foo\nnobar\nbaz')
        self.assertEqual(
            self.toggle_options_file.options,
            {'foo': True, 'bar': False, 'baz': True})

    def test_toggle_option_true(self):
        """Options can be enabled."""
        self.tempdir.mkfile(path=self.filename, content='foo\nnobar\nbaz')
        self.toggle_options_file.toggle('bar', True)
        # The value is written to file. For /proc and /sys files this will
        # result in passing the value to the kernel.
        self.assertEqual(
            self.readfile(self.tempdir.join(self.filename)), 'bar')

    def test_toggle_option_false(self):
        """Options can be disabled."""
        self.tempdir.mkfile(path=self.filename, content='foo\nnobar\nbaz')
        self.toggle_options_file.toggle('foo', False)
        # When an option is disabled the name is prefixed with 'no'.
        self.assertEqual(
            self.readfile(self.tempdir.join(self.filename)), 'nofoo')

    def test_toggle_option_unknown(self):
        """Passing an unknown option name raises a ValueError."""
        self.tempdir.mkfile(path=self.filename, content='foo\nnobar\nbaz')
        self.assertRaises(
            ValueError, self.toggle_options_file.toggle, 'other', True)


class ValueFileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.value_file = ValueFile(self.tempdir.join(self.filename))

    def test_value(self):
        """Value from the file can be returned."""
        self.tempdir.mkfile(path=self.filename, content='55\n')
        self.assertEqual(self.value_file.value, '55')

    def test_set(self):
        """Value can be set to file."""
        self.value_file.set('99')
        self.assertEqual(self.readfile(self.tempdir.join(self.filename)), '99')


class ToggleFileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.toggle_file = ToggleFile(self.tempdir.join(self.filename))

    def test_enabled_true(self):
        """If content is '1', the toggle is enabled."""
        self.tempdir.mkfile(path=self.filename, content='1\n')
        self.assertTrue(self.toggle_file.enabled)

    def test_enabled_false(self):
        """If content is '0', the toggle is not enabled."""
        self.tempdir.mkfile(path=self.filename, content='0\n')
        self.assertFalse(self.toggle_file.enabled)

    def test_toggle_true(self):
        """The value can be set to enabled."""
        self.toggle_file.toggle(True)
        self.assertTrue(self.toggle_file.enabled)
        self.assertEqual(self.readfile(self.tempdir.join(self.filename)), '1')

    def test_toggle_false(self):
        """The value can be set to disabled."""
        self.toggle_file.toggle(False)
        self.assertFalse(self.toggle_file.enabled)
        self.assertEqual(self.readfile(self.tempdir.join(self.filename)), '0')



import os

from ..testing import TestCase
from ..fs import (
    Entity,
    File,
    Directory)


class EntityTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'file'
        self.entity = Entity(self.tempdir.join(self.filename))

    def test_path(self):
        """The Entity path attribute contains the absolute path."""
        self.assertEqual(
            self.entity.path, os.path.join(self.tempdir.path, self.filename))

    def test_exists_false(self):
        """The exists property is False if the entity doesn't exist."""
        self.assertFalse(self.entity.exists)

    def test_exists_true(self):
        """The exists property is True if the entity exists."""
        self.tempdir.mkfile(path=self.filename)
        self.assertTrue(self.entity.exists)

    def test_readable_false(self):
        """The readable property is False if the path is not readable."""
        self.tempdir.mkfile(self.filename, mode=0o200)
        self.assertFalse(self.entity.readable)

    def test_readable_true(self):
        """The readable property is True if the path is readable."""
        self.tempdir.mkfile(path=self.filename)
        self.assertTrue(self.entity.readable)

    def test_writable_false(self):
        """The writable property is False if the path is not writable."""
        self.tempdir.mkfile(path=self.filename, mode=0o400)
        self.assertFalse(self.entity.writable)

    def test_writable_true(self):
        """The writable property is True if the path is writable."""
        self.tempdir.mkfile(path=self.filename)
        self.assertTrue(self.entity.writable)


class FileTests(TestCase):

    def setUp(self):
        super().setUp()
        self.filename = 'foo'
        self.file = File(os.path.join(self.tempdir.path, self.filename))

    def test_read(self):
        """File content can be read."""
        self.tempdir.mkfile(path=self.filename, content='some content')
        self.assertEqual(self.file.read(), 'some content')

    def test_write(self):
        """Content can be written to file."""
        self.file.write('some content')
        self.assertEqual(self.readfile(self.file.path), 'some content')


class DirectoryTests(TestCase):

    def setUp(self):
        super().setUp()
        self.dir = Directory(self.tempdir.path)
        self.dir.files = {'foo': File, 'bar': File}

    def test_list_specified(self):
        """Only names listed among Directory.files are returned."""
        self.tempdir.mkfile(path='foo')
        self.tempdir.mkfile(path='bar')
        self.tempdir.mkfile(path='bza')
        self.assertEqual(self.dir.list(), ['bar', 'foo'])

    def test_list_existing(self):
        """Only existing files are included in the listing."""
        self.tempdir.mkfile(path='foo')
        self.assertEqual(self.dir.list(), ['foo'])

    def test_listdir(self):
        """All names in the directory are returned."""
        self.tempdir.mkfile(path='foo')
        self.tempdir.mkfile(path='bar')
        self.assertCountEqual(self.dir.listdir(), ['bar', 'foo'])

    def test_iterable(self):
        """The Directory is iterable and returns Files in the directory."""
        self.tempdir.mkfile(path='foo')
        self.tempdir.mkfile(path='bar')
        file_list = list(self.dir)
        for elem in file_list:
            self.assertIsInstance(elem, File)
        self.assertEqual([elem.name for elem in file_list], ['bar', 'foo'])

    def test_get_file(self):
        """File items can be accessed."""
        self.tempdir.mkfile(path='foo', content='foo text')
        file_item = self.dir['foo']
        self.assertIsInstance(file_item, File)
        self.assertEqual(file_item.read(), 'foo text')

    def test_get_file_unknown(self):
        """Accessing an unknown file name raises an error."""
        self.assertRaises(KeyError, self.dir.__getitem__, 'unknown')

    def test_get_file_not_existing(self):
        """Accessing a known file that doesn't exist raises an error."""
        self.assertRaises(KeyError, self.dir.__getitem__, 'foo')

    def test_get_directory(self):
        """A Directory can contains sub-Directories."""

        class SubDirectory(Directory):
            files = {'foo': File}

        self.tempdir.mkfile(
            path=os.path.join('subdir', 'foo'), content='foo text')
        self.dir.files = {'subdir': SubDirectory}
        # The directory shows in the parent list
        self.assertEqual(self.dir.list(), ['subdir'])
        # The file is accessible through the tree
        self.assertEqual(self.dir['subdir']['foo'].read(), 'foo text')

    def test_join(self):
        """It's possible to join a path with the Directory one."""
        self.assertEqual(
            self.dir.join('append', 'path'),
            os.path.join(self.tempdir.path, 'append', 'path'))



from os import path

from ...testing import TestCase
from ..process import Process
from ..collection import (
    Collector,
    Collection)


class CollectorTests(TestCase):

    def test_collector_scan(self):
        """The Collector returns Processes for all PIDs."""
        self.make_process_file(10, 'cmdline')
        self.make_process_file(20, 'cmdline')
        self.make_process_file(30, 'cmdline')
        collector = Collector(proc=self.tempdir.path)
        self.assertCountEqual(
            (process.pid for process in collector.collect()),
            [10, 20, 30])

    def test_collector_with_pids(self):
        """If PIDs are provided, only those are included."""
        self.make_process_file(10, 'cmdline')
        self.make_process_file(20, 'cmdline')
        self.make_process_file(30, 'cmdline')
        collector = Collector(proc=self.tempdir.path, pids=(10, 30))
        self.assertCountEqual(
            (process.pid for process in collector.collect()),
            [10, 30])

    def test_collector_with_pids_sorts(self):
        """Processes are listed in PID order."""
        self.make_process_file(10, 'cmdline')
        self.make_process_file(20, 'cmdline')
        collector = Collector(proc=self.tempdir.path, pids=(20, 10))
        self.assertCountEqual(
            (process.pid for process in collector.collect()),
            [10, 20])

    def test_collector_skip_non_existing(self):
        """Collector skips non-existing processes."""
        self.make_process_file(10, 'cmdline')
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        self.assertEqual(
            [process.pid for process in collector.collect()], [10])


class CollectionTests(TestCase):

    def setUp(self):
        super().setUp()
        pids = (10, 30, 20)  # Not ordered so tests can sort
        self.collector = Collector(proc=self.tempdir.path, pids=pids)
        self.make_process_file(10, 'comm', content='foo')
        self.make_process_file(20, 'comm', content='zza')
        self.make_process_file(30, 'comm', content='bar')

    def process_list(self, pids):
        """Return a list of processes with specified pids."""
        return [
            Process(pid, path.join(self.tempdir.path, str(pid)))
            for pid in pids]

    def test_default_collector(self):
        """If a collector is not specified, one that looks at /proc is set."""
        collection = Collection()
        self.assertEqual(collection._collector._proc, '/proc/')

    def test_iter(self):
        """Collector is an iterable yielding Processes."""
        collection = Collection(collector=self.collector)
        self.assertCountEqual(collection, self.process_list([10, 20, 30]))

    def test_sort_by(self):
        """Collector can sort by the specified Process stat."""
        collection = Collection(collector=self.collector, sort_by='comm')
        self.assertEqual(list(collection), self.process_list([30, 10, 20]))

    def test_sort_by_reversed(self):
        """Collector can sort in reverse order."""
        collection = Collection(collector=self.collector, sort_by='-comm')
        self.assertEqual(list(collection), self.process_list([20, 10, 30]))

    def test_add_filter(self):
        """Collector.add_filter adds a filter for processes."""
        collection = Collection(collector=self.collector)
        collection.add_filter(lambda proc: proc.pid != 20)
        self.assertCountEqual(collection, self.process_list([10, 30]))

    def test_filter_multiple(self):
        """Multiple filters can be added."""
        collection = Collection(collector=self.collector)
        collection.add_filter(lambda proc: proc.pid != 20)
        collection.add_filter(lambda proc: proc.pid != 30)
        self.assertCountEqual(collection, self.process_list([10]))

    def test_filter_exclusive(self):
        """Filters are applied in 'or'."""
        collection = Collection(collector=self.collector)
        collection.add_filter(lambda proc: proc.pid == 10)
        collection.add_filter(lambda proc: proc.pid != 10)
        self.assertCountEqual(collection, [])



"""
Handle collection of processes, allowing filtering and sorting by attribute
values.
"""

from os import path
from glob import iglob

from .process import Process


class Collector:
    """Process collector.

    An iterable with PIDs can be passed, otherwise all PIDs found scanning
    the ``/proc`` directory are returned.

    """
    def __init__(self, proc='/proc', pids=()):
        self._proc = '{}/'.format(path.normpath(proc))
        self._pids = pids

    def collect(self):
        """Return an iterator yielding Process objects."""
        if self._pids:
            proc_dirs = (
                '{}{}'.format(self._proc, pid) for pid in sorted(self._pids))
        else:
            proc_dirs = iglob('{}[0-9]*'.format(self._proc))

        l = len(self._proc)
        for proc_dir in proc_dirs:
            process = Process(int(proc_dir[l:]), proc_dir)
            process.collect_stats()
            if process.exists:
                # Don't return non-existing processes. Check this after trying
                # collecting stats, since doing the opposite there's a chance
                # the process might go away bewteen the check and the
                # data collection.
                yield process


class Collection:
    """A Process collection.

    Parameters:
      collector: A ProcessCollector instance. If not provided, a default one
        will be created.
      sort_by: The field to sort processes by. It can be prefixed with ``-``
        to invert sorting (e.g. ``pid`` or ``-pid``).

    """

    def __init__(self, collector=None, sort_by=None):
        if collector is None:
            self._collector = Collector()
        else:
            self._collector = collector
        self._filters = []
        self._set_sort_by(sort_by)

    def add_filter(self, filter_function):
        """Add a filtering function to the collection.

        Processes not matching the filter are not returned.

        Parameters:
          filter_function: A callable accepting a Process and returning a
            boolean value (whether the process should be included).

        """
        self._filters.append(filter_function)

    def __iter__(self):
        """Return an iterator yielding Process objects.

        Processes are filtered and sorted as configured.

        """
        iterator = self._collector.collect()
        if self._filters:
            iterator = filter(self._filter, iterator)

        if self._sort_by is not None:
            def key(elem):
                return elem.get(self._sort_by)

            iterator = iter(sorted(
                iterator, key=key, reverse=self._sort_reverse))

        return iterator

    def _filter(self, proc):
        """Apply filters to a Process."""
        return all(ffunc(proc) for ffunc in self._filters)

    def _set_sort_by(self, sort_by):
        if sort_by is not None and sort_by.startswith('-'):
            self._sort_by = sort_by[1:]
            self._sort_reverse = True
        else:
            self._sort_by = sort_by
            self._sort_reverse = False



"""Files and directories access.

The :class:`File` and :class:`Directory` classes provide abstractions to access
filesytem entities.

"""

import os


class Entity:
    """A filesystem entity such as a file or directory."""

    def __init__(self, path):
        self.path = path
        self.name = os.path.basename(self.path)

    @property
    def exists(self):
        """Whether the path exists."""
        return os.path.exists(self.path)

    @property
    def readable(self):
        """Whether the path is readable."""
        return os.access(self.path, os.R_OK)

    @property
    def writable(self):
        """Whether the path is writable."""
        return os.access(self.path, os.W_OK)


class File(Entity):
    """Wrapper to reaad/write a file."""

    def read(self):
        """Return file content."""
        with open(self.path) as fh:
            return fh.read()

    def write(self, content):
        """Write content to file, replacing the content if it exists."""
        with open(self.path, 'w') as fh:
            fh.write(content)


class Directory(Entity):
    """Access  files in a directory with a :class:`dict`-like interface."""

    #: Map names of files under the directory to their corresponding
    #: :class:`File` type.  Subclasses should define this.
    files = {}

    def list(self):
        """Return a list of names in the directory.

        Only existing files that match names listed in `files` are returned.

        """
        return sorted(
            name for name in self.files if os.path.exists(self._path(name)))

    def listdir(self):
        """Return all existing names in a directory."""
        return os.listdir(self.path)

    def join(self, *paths):
        """Append the given path to the directory one."""
        return os.path.join(self.path, *paths)

    def __getitem__(self, name):
        """Return the :class:`File` instance for a name."""
        path = self._path(name)
        item = self.files[name](path)
        if not item.exists:
            raise KeyError(name)
        return item

    def __iter__(self):
        """Return an iterator yielding :class:`File`s in the directory."""
        for name in self.list():
            yield self[name]

    def _path(self, name):
        return os.path.join(self.path, name)



"""Parsers for :file:`/proc` files containing system information."""

import re

from ..text import (
    ParsedFile,
    SingleLineFile)


class ProcStat(ParsedFile):
    """Parse :file:`/proc/stat`."""

    stat_fields = [
        'user', 'nice', 'system', 'idle', 'iowait', 'irq', 'softirq', 'steal',
        'guest', 'guest-nice']

    def _parse(self, content):
        result = {}

        for line in content.splitlines():
            if not line.startswith('cpu'):
                # Only CPU stats are reported for now.
                break

            values = line.split()
            label = values.pop(0)
            values = [float(value) for value in values]
            total = sum(values)
            values = [(value / total) for value in values]
            # If there are less fields than the declared ones, they are ignored
            # by zip().
            result[label] = dict(zip(self.stat_fields, values))
        return result


class ProcUptime(SingleLineFile):
    """Parse :file:`/proc/uptime`."""

    fields = (('uptime', float), ('idle', float))


class ProcLoadavg(SingleLineFile):
    """Parse :file:`/proc/loadavg`."""

    fields = (('load1', float), ('load5', float), ('load15', float))


class ProcVmstat(ParsedFile):
    """Parse :file:`/proc/vmstat`."""

    def _parse(self, content):
        items = (line.split() for line in content.splitlines())
        return dict((key, int(value)) for key, value in items)


class ProcDiskstats(ParsedFile):
    """Parse :file:`/proc/diskstats`."""

    diskstat_fields = [
        'read', 'read-merged', 'read-sect', 'read-ms', 'write', 'write-merged',
        'write-sect', 'write-ms', 'io-curr', 'io-ms', 'io-ms-weighted']

    def _parse(self, content):
        result = {}
        for line in content.splitlines():
            split = line.split()[2:]  # Ignore major/minor fields
            dev_name, values = split[0], split[1:]
            values = [int(value) for value in values]
            result[dev_name] = dict(zip(self.diskstat_fields, values))
        return result


class ProcMeminfo(ParsedFile):
    """Parse :file:`/proc/meminfo`."""

    _parse_re = re.compile(r'(?P<name>.+):\s+(?P<value>[0-9]+)')

    def _parse(self, content):
        return dict(self._parse_line(line) for line in content.splitlines())

    def _parse_line(self, line):
        match = self._parse_re.match(line).groupdict()
        return match['name'], int(match['value'])


class ProcCgroups(ParsedFile):
    """Parse :file:`/proc/cgroups`."""

    def _parse(self, content):
        result = {}
        for line in content.splitlines():
            if line.startswith('#'):
                continue
            subsys, hier_id, num_cgroups, enabled = line.split()
            result[subsys] = {
                'hierarchy-id': int(hier_id),
                'num-cgroups': int(num_cgroups),
                'enabled': enabled == '1'}
        return result



from io import StringIO

from ....testing import TestCase
from .. import TableFormatter
from ...collection import (
    Collector,
    Collection)


class TableFormatterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.stream = StringIO()
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        self.collection = Collection(collector=collector)
        self.make_process_file(10, 'cmdline', content='/bin/foo')
        self.make_process_file(20, 'cmdline', content='/bin/bar')

    def test_format(self):
        """TableFormatter outputs a table with processes data."""
        formatter = TableFormatter(self.stream, ['pid', 'cmd'])
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            ' pid  cmd      \n'
            ' 10   /bin/foo \n'
            ' 20   /bin/bar \n')

    def test_format_border(self):
        """TableFormatter can add borders to the table."""
        formatter = TableFormatter(
            self.stream, ['pid', 'cmd'], borders=True)
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            '+-----+----------+\n'
            '| pid | cmd      |\n'
            '+-----+----------+\n'
            '| 10  | /bin/foo |\n'
            '| 20  | /bin/bar |\n'
            '+-----+----------+\n')



"""Access system and process statistics through files under :file:`/proc`.

System statistics, such as load, uptime, memory and disk information can be
read through the :class:`ProcDirectory`. For example::

  >>> ProcDirectory('/proc')['uptime'].parse()
  {'uptime': 695283.06, 'idle': 1376159.44}

Process-specific stats, like used memory, IO, etc. can be accessed through
:class:`ProcProcessDirectory`, such as::

  >>> ProcProcessDirectory('/proc/self')['io'].parse()
  {'rchar': 405786, 'syscr': 415, 'cancelled_write_bytes': 0, 'syscw': 256,
   'write_bytes': 20480, 'read_bytes': 32768, 'wchar': 14790}

"""

from ...fs import Directory
from ..types import ValueFile
from .system import (
    ProcDiskstats,
    ProcLoadavg,
    ProcMeminfo,
    ProcStat,
    ProcVmstat,
    ProcUptime,
    ProcCgroups)
from .process import (
    ProcPIDCmdline,
    ProcPIDIo,
    ProcPIDStat,
    ProcPIDStatm,
    ProcPIDStatus,
    ProcPIDEnviron,
    ProcPIDSched,
    ProcPIDCgroup)


class ProcDirectory(Directory):
    """The :file:`/proc` directory."""

    files = {
        'cgroups': ProcCgroups,
        'diskstats': ProcDiskstats,
        'loadavg': ProcLoadavg,
        'meminfo': ProcMeminfo,
        'vmstat': ProcVmstat,
        'stat': ProcStat,
        'uptime': ProcUptime,
        'vmstat': ProcVmstat}


class ProcProcessDirectory(Directory):
    """A directory for a process under :file:`/proc/[pid]`."""

    files = {
        'cgroup': ProcPIDCgroup,
        'cmdline': ProcPIDCmdline,
        'comm': ValueFile,
        'environ': ProcPIDEnviron,
        'io': ProcPIDIo,
        'sched': ProcPIDSched,
        'stat': ProcPIDStat,
        'statm': ProcPIDStatm,
        'status': ProcPIDStatus,
        'task': Directory,
        'wchan': ValueFile}


class ProcTaskDirectory(Directory):
    """A directory for a task under :file:`/proc/[pid]/task/[tid]`."""

    files = {
        'cgroup': ProcPIDCgroup,
        'cmdline': ProcPIDCmdline,
        'comm': ValueFile,
        'environ': ProcPIDEnviron,
        'io': ProcPIDIo,
        'sched': ProcPIDSched,
        'stat': ProcPIDStat,
        'statm': ProcPIDStatm,
        'wchan': ValueFile}



import json
from io import StringIO

from ....testing import TestCase
from .. import JSONFormatter
from ...collection import (
    Collector,
    Collection)


class JSONFormatterTests(TestCase):

    def setUp(self):
        super().setUp()
        self.stream = StringIO()
        collector = Collector(proc=self.tempdir.path, pids=(10, 20))
        self.collection = Collection(collector=collector)
        self.make_process_file(10, 'cmdline', content='/bin/foo')
        self.make_process_file(20, 'cmdline', content='/bin/bar')

    def test_format(self):
        """JSONFormatter formats process info in JSON format."""
        formatter = JSONFormatter(self.stream, ['pid', 'cmd'])
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            json.dumps(
                {'fields': ['pid', 'cmd'],
                 'processes': [
                     {'pid': 10, 'cmd': '/bin/foo'},
                     {'pid': 20, 'cmd': '/bin/bar'}]}))

    def test_format_indent(self):
        """The indent parameter is passed to the JSON encoder."""
        formatter = JSONFormatter(self.stream, ['pid', 'cmd'], indent=3)
        formatter.format(self.collection)
        self.assertEqual(
            self.stream.getvalue(),
            json.dumps(
                {'fields': ['pid', 'cmd'],
                 'processes': [
                     {'pid': 10, 'cmd': '/bin/foo'},
                     {'pid': 20, 'cmd': '/bin/bar'}]},
                indent=3))



from ...testing import TestCase
from ..text import (
    ParsedFile,
    SingleLineFile,
    SplittedFile)


class ParsedFileTests(TestCase):

    def test_parse_not_implemented(self):
        """Parsedfile._parse must be implemented by subclasses."""
        path = self.tempdir.mkfile()
        parsed_file = ParsedFile(path)
        self.assertRaises(NotImplementedError, parsed_file.parse)

    def test_parse(self):
        """ParsedFile.parse calls the parser with the file content."""
        content = 'line 1\nline 2'
        path = self.tempdir.mkfile(content=content)
        parsed_file = ParsedFile(path)
        parsed_file._parse = lambda content: content
        self.assertEqual(parsed_file.parse(), content)

    def test_parse_not_existent(self):
        """FileParser.parse returns None if file doesn't exist."""
        path = self.tempdir.join('file')
        parsed_file = ParsedFile(path)
        # The parser method is not called
        self.assertIsNone(parsed_file.parse())


class SingleLineFileTests(TestCase):

    def test_parse_empty_file(self):
        """If the file is empty, an empty list is returned."""
        path = self.tempdir.mkfile()
        single_line_file = SingleLineFile(path)
        self.assertEqual(single_line_file.parse(), [])

    def test_parse_empty_file_with_fields(self):
        """If the file is empty and fields are set, return an empty dict."""
        path = self.tempdir.mkfile()
        single_line_file = SingleLineFile(path)
        single_line_file.fields = ('foo', 'bar')
        self.assertEqual(single_line_file.parse(), {})

    def test_parse_only_first_line(self):
        """Only the first line of the file is parsed."""
        path = self.tempdir.mkfile(content='foo\nbar\nbaz\n')
        single_line_file = SingleLineFile(path)
        self.assertEqual(single_line_file.parse(), ['foo'])

    def test_parse_default_separator(self):
        """By default the file content is split on spaces."""
        path = self.tempdir.mkfile(content='foo bar baz')
        single_line_file = SingleLineFile(path)
        self.assertEqual(single_line_file.parse(), ['foo', 'bar', 'baz'])

    def test_parse_different_separator(self):
        """The file content is split on a custom separator."""
        path = self.tempdir.mkfile(content='foo|bar|baz')
        single_line_file = SingleLineFile(path)
        single_line_file.separator = '|'
        self.assertEqual(single_line_file.parse(), ['foo', 'bar', 'baz'])

    def test_parse_no_separator(self):
        """If no separator is set, the stripped content is returned."""
        path = self.tempdir.mkfile(content='just some text\n')
        single_line_file = SingleLineFile(path)
        single_line_file.separator = None
        self.assertEqual(single_line_file.parse(), 'just some text')

    def test_parse_with_fields(self):
        """If fields are specified, a dict is returned upon parsing."""
        path = self.tempdir.mkfile(content='foo bar baz')
        single_line_file = SingleLineFile(path)
        single_line_file.fields = ('one', 'two', 'three')
        self.assertEqual(
            single_line_file.parse(),
            {'one': 'foo', 'two': 'bar', 'three': 'baz'})

    def test_parse_with_fields_with_type(self):
        """Typed fields are converted in the result."""
        path = self.tempdir.mkfile(content='foo 1 30.3')
        single_line_file = SingleLineFile(path)
        single_line_file.fields = ('one', ('two', int), ('three', float))
        self.assertEqual(
            single_line_file.parse(), {'one': 'foo', 'two': 1, 'three': 30.3})

    def test_parse_with_fields_with_none(self):
        """If a field is None, it's skipped in the result."""
        path = self.tempdir.mkfile(content='foo baz bar')
        single_line_file = SingleLineFile(path)
        single_line_file.fields = ('one', None, 'three')
        self.assertEqual(
            single_line_file.parse(), {'one': 'foo', 'three': 'bar'})

    def test_parse_with_less_fields(self):
        """If fields less fields are present, they're skipped in the result."""
        path = self.tempdir.mkfile(content='foo bar')
        single_line_file = SingleLineFile(path)
        single_line_file.fields = ('one', 'two', 'three', 'four')
        self.assertEqual(
            single_line_file.parse(), {'one': 'foo', 'two': 'bar'})


class SplittedFileTests(TestCase):

    def test_parse_single_line(self):
        """A single line of content is split on spaces."""
        path = self.tempdir.mkfile(content='foo bar baz')
        splitted_file = SplittedFile(path)
        self.assertEqual(splitted_file.parse(), ['foo', 'bar', 'baz'])

    def test_parse_multiple_lines(self):
        """A file with multiple lines is split on newlines."""
        path = self.tempdir.mkfile(content='foo\nbar\nbaz')
        splitted_file = SplittedFile(path)
        self.assertEqual(splitted_file.parse(), ['foo', 'bar', 'baz'])

    def test_parse_multiple_lines_trailing_newline(self):
        """Trailing newline is ignored."""
        path = self.tempdir.mkfile(content='foo\nbar\nbaz\n')
        splitted_file = SplittedFile(path)
        self.assertEqual(splitted_file.parse(), ['foo', 'bar', 'baz'])



"""Hold information about a running process.

Stats for a process can be collected and accessed::

p = Process(123, '/proc/123')
p.collect_stats()
p.get('statm.size')

"""

from datetime import datetime

from ..files.proc import ProcProcessDirectory


class TaskBase:
    """Base class for tasks and processes."""

    _utcnow = datetime.utcnow  # For testing

    _id_attr = '_id'

    def __init__(self, id, proc_dir):
        self._id = id
        self._dir = ProcProcessDirectory(proc_dir)
        self._reset()

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self._id)

    def __eq__(self, other):
        return self._id == other._id

    def __hash__(self):
        return hash((self.__class__, self._id))

    @property
    def cmd(self):
        """The task command line, with brackets for kernel tasks."""
        cmdline = self._stats.get('cmdline')
        if cmdline:
            return ' '.join(cmdline)
        comm = self._stats.get('comm')
        return '[{}]'.format(comm) if comm else ''

    @property
    def exists(self):
        """Whether the task exists."""
        return self._dir.exists

    def collect_stats(self):
        """Collect stats about the process from ``/proc`` files."""
        self._reset()

        if not self._dir.readable:
            return

        self._timestamp = self._utcnow()

        for name in self._dir.list():
            entry = self._dir[name]
            if not entry.readable or not hasattr(entry, 'parse'):
                continue

            try:
                parsed_stats = entry.parse()
            except IOError:
                continue

            if isinstance(parsed_stats, dict):
                self._stats.update(
                    ('{}.{}'.format(name, key), value)
                    for key, value in parsed_stats.items())
            else:
                self._stats[name] = parsed_stats

    def available_stats(self):
        """Return a sorted list of available stats for the process."""
        return sorted(self._stats)

    def stats(self):
        """Return a dict with process stats."""
        return self._stats.copy()

    @property
    def timestamp(self):
        """Return the timestamp for stat collection."""
        return self._timestamp

    def get(self, stat):
        """Return the stat with the name, or None if not available."""
        if stat in (self._id_attr, 'cmd', 'timestamp'):
            return getattr(self, stat)

        return self._stats.get(stat)

    def _reset(self):
        """Reset stats."""
        self._stats = {}
        self._timestamp = None


class Process(TaskBase):
    """Retrieve and hold information about a given process."""

    _utcnow = datetime.utcnow  # For testing

    _id_attr = 'pid'

    @property
    def pid(self):
        """The process PID."""
        return self._id

    def tasks(self):
        """Return a list of Tasks for the Process."""
        tasks = []
        tasks_dir = self._dir['task']
        for tid in tasks_dir.listdir():
            tasks.append(Task(int(tid), self, tasks_dir.join(tid)))
        return tasks


class Task(TaskBase):
    """Retrieve and hold information about a given task."""

    _id_attr = 'tid'

    def __init__(self, id, parent, proc_dir):
        super().__init__(id, proc_dir)
        self.parent = parent

    @property
    def tid(self):
        """The task TID."""
        return self._id



from unittest import TestCase

from .. import (
    get_formats,
    get_formatter)
from ..fmt_table import TableFormatter


class GetFormatsTest(TestCase):

    def test_get_formats(self):
        """get_formats return a sorted list of formatters."""
        self.assertEqual(get_formats(), ['csv', 'json', 'table'])


class GetFormatterTest(TestCase):

    def test_get_formatter(self):
        """get_formatter return a formatter by name."""
        self.assertIs(get_formatter('table'), TableFormatter)



"""Access files under the :file:`/sys` filesytem.

This module currently allows to access and configure tracing options via
:class:`TracingDirectory`.

"""

from ...fs import (
    Directory,
    File)
from ..types import (
    OptionsFile,
    ValueFile,
    ToggleFile,
    SelectableOptionsFile,
    TogglableOptionsFile)


class TracingDirectory(Directory):
    """A :file:`/sys/kernel/debug/tracing/instance/[tracer]` directory."""

    files = {
        'available_tracers': OptionsFile,
        'current_tracer': ValueFile,
        'trace_clock': SelectableOptionsFile,
        'trace_options': TogglableOptionsFile,
        'trace_marker': File,
        'tracing_on': ToggleFile
    }



"""Read and write files under :file:`/proc` and :file:`/sys` Linux filesystems.

This module and its submoduels contain specialized classes to read, parse and
write content to different types of files under :file:`/proc` and :file:`/sys`
Linux filesystems.

Submodules are structured as follows:

 - :mod:`lxstats.files.text`: base classes for reading and writing text files.

 - :mod:`lxstats.files.types`: classes for reading and writing different types
   of files, mostly used in the :file:`/sys` filesystem.

 - :mod:`lxstats.files.proc`: classes for accessing system-wide and per-process
   statistics.

 - :mod:`lxstats.files.sys`: classes for accessing files under :file:`/sys`,
   such as debug tracing.

"""



from textwrap import dedent

from ....testing import TestCase
from ..process import (
    ProcPIDCmdline,
    ProcPIDStat,
    ProcPIDStatm,
    ProcPIDIo,
    ProcPIDSched,
    ProcPIDEnviron,
    ProcPIDCgroup,
    ProcPIDStatus)


class ProcPIDCmdlineTests(TestCase):

    def test_parse(self):
        """Command line tokens are parsed and split into a list."""
        path = self.tempdir.mkfile(content='/bin/foo\x00bar\x00baz\x00')
        cmdline_file = ProcPIDCmdline(path)
        self.assertEqual(cmdline_file.parse(), ['/bin/foo', 'bar', 'baz'])

    def test_parse_with_spaces(self):
        """Command or arguments can contain spaces."""
        path = self.tempdir.mkfile(content='/bin/foo bar\x00baz bza\x00')
        cmdline_file = ProcPIDCmdline(path)
        self.assertEqual(cmdline_file.parse(), ['/bin/foo bar', 'baz bza'])


class ProcPIDStatTests(TestCase):

    def test_fields(self):
        """Fields and values from /proc/[pid]/stat files are reported."""
        content = ' '.join(str(i) for i in range(45))
        path = self.tempdir.mkfile(content=content)
        stat_file = ProcPIDStat(path)
        self.assertEqual(
            stat_file.parse(),
            {'pid': 0,
             'comm': '1',
             'state': '2',
             'ppid': 3,
             'pgrp': 4,
             'session': 5,
             'tty_nr': 6,
             'tpgid': 7,
             'flags': 8,
             'minflt': 9,
             'cminflt': 10,
             'majflt': 11,
             'cmajflt': 12,
             'utime': 13,
             'stime': 14,
             'cutime': 15,
             'cstime': 16,
             'priority': 17,
             'nice': 18,
             'num_threads': 19,
             'itrealvalue': 20,
             'starttime': 21,
             'vsize': 22,
             'rss': 23,
             'rsslim': 24,
             'startcode': 25,
             'endcode': 26,
             'startstack': 27,
             'kstkesp': 28,
             'kstkeip': 29,
             'exit_signal': 37,
             'processor': 38,
             'rt_priority': 39,
             'policy': 40,
             'delayacct_blkio_ticks': 41,
             'guest_time': 42,
             'cguest_time': 43})

    def test_comm_with_spaces(self):
        """The comm field can contain spaces.."""
        fields = [str(i) for i in range(45)]
        # Set a comm field with spaces
        fields[1] = '(cmd with spaces)'
        content = ' '.join(fields)
        path = self.tempdir.mkfile(content=content)
        stat_file = ProcPIDStat(path)
        stats = stat_file.parse()
        self.assertEqual(stats['comm'], 'cmd with spaces')


class ProcPIDStatmTests(TestCase):

    def test_fields(self):
        """Fields and values from /proc/[pid]/statm files are reported."""
        path = self.tempdir.mkfile(content='1 2 3 4 5 6 7')
        statm_file = ProcPIDStatm(path)
        self.assertEqual(
            statm_file.parse(),
            {'size': 1, 'resident': 2, 'share': 3, 'text': 4, 'lib': 5,
             'data': 6, 'dt': 7})


class ProcPIDIOTests(TestCase):

    def test_fields(self):
        """Fields and values from /proc/[pid]/io files are reported."""
        content = dedent(
            '''\
            rchar: 100
            wchar: 200
            syscr: 300
            syscw: 400
            read_bytes: 500
            write_bytes: 600
            cancelled_write_bytes: 700
            ''')
        path = self.tempdir.mkfile(content=content)
        io_file = ProcPIDIo(path)
        self.assertEqual(
            io_file.parse(),
            {'rchar': 100,
             'wchar': 200,
             'syscr': 300,
             'syscw': 400,
             'read_bytes': 500,
             'write_bytes': 600,
             'cancelled_write_bytes': 700})


class ProcPIDSchedTests(TestCase):

    def test_fields(self):
        """Fields and values from /proc/[pid]/sched files are parsed."""
        content = dedent(
            '''\
            process (1234, #threads: 1)
            -------------------------------------------------------------------
            se.exec_start                                :     123456789.123456
            se.vruntime                                  :       1234567.123456
            se.sum_exec_runtime                          :             0.123456
            se.statistics.sum_sleep_runtime              :             1.234567
            current_node=0, numa_group_id=0
            numa_faults node=0 task_private=0 task_shared=0 group_private=0
            ''')
        path = self.tempdir.mkfile(content=content)
        io_file = ProcPIDSched(path)
        self.assertEqual(
            io_file.parse(),
            {'se.exec_start': 123456789.123456,
             'se.vruntime': 1234567.123456,
             'se.sum_exec_runtime': 0.123456,
             'se.statistics.sum_sleep_runtime': 1.234567})


class ProcPIDEnvironTests(TestCase):

    def test_parse(self):
        """Environment variables are returned as a dict."""
        path = self.tempdir.mkfile(content='FOO=foo\x00BAR=bar\x00')
        environ_file = ProcPIDEnviron(path)
        self.assertEqual(environ_file.parse(), {'FOO': 'foo', 'BAR': 'bar'})

    def test_parse_single_value(self):
        """If the = sign is not present, the line is used as key."""
        path = self.tempdir.mkfile(content='someline\x00BAR=bar\x00')
        environ_file = ProcPIDEnviron(path)
        self.assertEqual(
            environ_file.parse(), {'someline': None, 'BAR': 'bar'})


class ProcPIDCgroupTests(TestCase):

    def test_parse(self):
        """A dict is returned with hierarchy id as key."""
        content = dedent(
            '''\
            6:hugetlb:/group2
            5:net_cls,net_prio:/group1
            4:blkio:/group2
            3:cpu,cpuacct:/group1
            2:devices:/group1
            ''')
        path = self.tempdir.mkfile(content=content)
        cgroup_file = ProcPIDCgroup(path)
        self.assertEqual(
            cgroup_file.parse(),
            {6: (['hugetlb'], '/group2'),
             5: (['net_cls', 'net_prio'], '/group1'),
             4: (['blkio'], '/group2'),
             3: (['cpu', 'cpuacct'], '/group1'),
             2: (['devices'], '/group1')})


class ProcPIDStatusTests(TestCase):

    def test_parse(self):
        """A dict with memory information for the file is returned."""
        content = dedent(
            '''\
            VmPeak:\t 1132616 kB
            VmSize:\t 1115196 kB
            VmLck:\t       0 kB
            VmPin:\t       0 kB
            VmHWM:\t  246340 kB
            VmRSS:\t  246340 kB
            RssAnon:\t  210612 kB
            RssFile:\t   35700 kB
            RssShmem:\t      28 kB
            VmData:\t  292668 kB
            VmStk:\t     548 kB
            VmExe:\t    2436 kB
            VmLib:\t  136932 kB
            VmPTE:\t    1272 kB
            VmPMD:\t      16 kB
            VmSwap:\t       0 kB
            HugetlbPages:\t       0 kB
            ''')
        path = self.tempdir.mkfile(content=content)
        status_file = ProcPIDStatus(path)
        self.assertEqual(
            status_file.parse(),
            {'VmPeak': 1159798784,
             'VmSize': 1141960704,
             'VmLck': 0,
             'VmPin': 0,
             'VmHWM': 252252160,
             'VmRSS': 252252160,
             'RssAnon': 215666688,
             'RssFile': 36556800,
             'RssShmem': 28672,
             'VmData': 299692032,
             'VmStk': 561152,
             'VmExe': 2494464,
             'VmLib': 140218368,
             'VmPTE': 1302528,
             'VmPMD': 16384,
             'VmSwap': 0,
             'HugetlbPages': 0})

    def test_parse_skip_extra(self):
        """Non memory-related info is skipped."""
        content = dedent(
            '''\
            Uid:\t1000	1000	1000	1000
            Gid:\t1000	1000	1000	1000
            VmPeak:\t 1132616 kB
            VmSize:\t 1115196 kB
            ''')
        path = self.tempdir.mkfile(content=content)
        status_file = ProcPIDStatus(path)
        self.assertEqual(
            status_file.parse(),
            {'VmPeak': 1159798784,
             'VmSize': 1141960704})

    def test_parse_split_in_two(self):
        """Lines containing more than one ':' are split in two."""
        content = dedent(
            '''\
            VmPeak:\t 1132616 kB
            Line: with more: semicolons
            VmSize:\t 1115196 kB
            ''')
        path = self.tempdir.mkfile(content=content)
        status_file = ProcPIDStatus(path)
        self.assertEqual(
            status_file.parse(),
            {'VmPeak': 1159798784,
             'VmSize': 1141960704})



"""Parse types of files used in /proc and /sys filesytems."""


from collections import OrderedDict

from .text import (
    SplittedFile,
    SingleLineFile)


class OptionsFile(SplittedFile):
    """File listing a set of options.

    It returns available options for a file, including the selected one, if
    present.

    For example, for a file containing::

      foo [bar] baz

    :meth:`options` returns::

      ['foo', 'bar', 'baz']

    """

    @property
    def options(self):
        """Return a list with avalilable options."""
        return [self._strip_selected(value) for value in self.parse()]

    def _strip_selected(self, value):
        return value[1:-1] if value.startswith('[') else value


class SelectableOptionsFile(OptionsFile):
    """File listing a set of options with a single selected one.


    It works like :class:`OptionsFile`, but also allow getting and setting the
    selected option.

    """

    @property
    def selected(self):
        """Return the selected option."""
        for value in self.parse():
            if value.startswith('['):
                return self._strip_selected(value)

    def select(self, value):
        """Set the specified option value.

        :class:`ValueError` is raised if the value is not valid."""

        if value not in self.options:
            raise ValueError(value)
        self.write(value)


class TogglableOptionsFile(OptionsFile):
    """A file with a list of options that can be individually enabled.

    Disabled options have names prefixed by :data:`no`.

    For example, for a file containing::

      foo
      nobar
      baz

    :meth:`options` returns::

      {'foo': True,
       'bar': False,
       'baz': True}

    """

    @property
    def options(self):
        """Return a dict with options and their current values."""
        options = OrderedDict()
        for option in super().options:
            value = not option.startswith('no')
            if not value:
                option = option[2:]
            options[option] = value

        return options

    def toggle(self, option, value):
        """Change the value of the specified option."""
        if option not in self.options:
            raise ValueError(option)

        prefix = '' if value else 'no'
        self.write('{}{}'.format(prefix, option))


class ValueFile(SingleLineFile):
    """File containing a single value that can be read or set."""

    separator = None

    @property
    def value(self):
        """Return the current value in the file."""
        return self.parse()

    def set(self, value):
        self.write(value)


class ToggleFile(SingleLineFile):
    """File to enable or disable an option.

    The file contains a boolean value expressed as :data:`0` or :data:`1`.

    """

    separator = None

    @property
    def enabled(self):
        """Return whether the toggle value is enabled."""
        return self.parse() == '1'

    def toggle(self, value):
        """Enable or disable the value based on the passed :class:`bool`."""
        content = '1' if value else '0'
        self.write(content)



import os
from unittest import mock
from datetime import datetime

from ...testing import TestCase
from ..process import (
    TaskBase,
    Process,
    Task)


class TaskBaseTests(TestCase):

    def setUp(self):
        super().setUp()
        self.id = 10
        self.task = TaskBase(
            self.id, os.path.join(self.tempdir.path, str(self.id)))

    def test_repr(self):
        """__repr__ includes the task ID."""
        self.assertEqual(repr(self.task), 'TaskBase(10)')

    def test_id(self):
        """The id attribute returns the task identifier."""
        self.assertEqual(self.task._id, self.id)

    def test_exists(self):
        """It's possible to check whether a process exists."""
        self.assertFalse(self.task.exists)
        self.make_process_file(self.id, 'cmdline', content='cmd')
        self.assertTrue(self.task.exists)

    def test_collect_stats(self):
        """Stats are collected from proc when collect_stats is called."""
        self.assertEqual([], self.task.available_stats())
        self.make_process_file(self.id, 'cmdline', content='cmd')
        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), ['cmdline'])

    def test_collect_stats_no_proc_dir(self):
        """If the task dir is not found, stats are left empty."""
        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), [])
        self.assertEqual(self.task.stats(), {})

    def test_collect_stats_prefix(self):
        """The file prefix is used if it reports multiple stats."""
        self.make_process_file(self.id, 'statm', content='1 2 3 4 5 6 7')
        self.task.collect_stats()
        self.assertEqual(
            self.task.stats(),
            {'statm.size': 1, 'statm.resident': 2, 'statm.share': 3,
             'statm.text': 4, 'statm.lib': 5, 'statm.data': 6, 'statm.dt': 7})

    def test_collect_stats_unreadable_file(self):
        """Unreadable files are skipped when reading stats."""
        self.make_process_file(self.id, 'cmdline', content='cmd', mode=0o200)
        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), [])

    def test_collect_stats_not_parsable(self):
        """STats are not collected for entries that are not parsable files."""
        self.make_process_dir(self.id, 'task')
        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), [])

    @mock.patch('lxstats.files.text.ParsedFile.parse')
    def test_collect_stats_ioerror(self, mock_file):
        """If reading a file raises an IOError, the stat is skipped."""
        self.make_process_file(self.id, 'cmdline', content='cmd')
        mock_file.side_effect = IOError()

        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), [])

    def test_cmd_from_cmdline(self):
        """TaskBase.cmd parses the content of cmdline if not empty."""
        self.make_process_file(
            self.id, 'cmdline', content='cmd\x00with\x00args\x00')
        self.task.collect_stats()
        self.assertNotIn('comm', self.task.available_stats())
        self.assertEqual(
            self.task.get('cmdline'), ['cmd', 'with', 'args'])
        self.assertEqual(self.task.cmd, 'cmd with args')

    def test_cmd_from_comm(self):
        """If cmdline is empty, comm is read."""
        self.make_process_file(self.id, 'comm', content='cmd')
        self.task.collect_stats()
        self.assertNotIn('cmdline', self.task.available_stats())
        self.assertEqual(self.task.cmd, '[cmd]')

    def test_cmd_empty(self):
        """If cmdline and cmd are not found, an empty string is returned."""
        self.task.collect_stats()
        self.assertEqual(self.task.cmd, '')

    def test_timestamp_empty(self):
        """The timestamp is None when stats have not been collected."""
        self.make_process_file(self.id, 'cmdline', content='cmd')
        self.assertIsNone(self.task.timestamp)

    def test_timestamp(self):
        """The timestamp is None when stats have not been collected."""
        self.make_process_file(self.id, 'cmdline', content='cmd')
        now = datetime.utcnow()
        self.task._utcnow = lambda: now
        self.task.collect_stats()
        self.assertEqual(self.task.timestamp, now)

    def test_available_stats(self):
        """Available Process stats can be listed."""
        self.make_process_file(self.id, 'comm', content='cmd')
        self.make_process_file(self.id, 'wchan', content='0')
        self.task.collect_stats()
        self.assertEqual(self.task.available_stats(), ['comm', 'wchan'])

    def test_stats(self):
        """Available Process  stats can be returned as a dict ."""
        self.make_process_file(self.id, 'comm', content='cmd')
        self.make_process_file(self.id, 'wchan', content='0')
        self.task.collect_stats()
        self.assertEqual(self.task.stats(), {'comm': 'cmd', 'wchan': '0'})

    def test_get(self):
        """Value for a stat can be returned."""
        self.make_process_file(
            self.id, 'wchan', content='poll_schedule_timeout')
        self.task.collect_stats()
        self.assertEqual(self.task.get('wchan'), 'poll_schedule_timeout')

    def test_get_not_found(self):
        """If the requested stat is not found, None is returned."""
        self.task.collect_stats()
        self.assertIsNone(self.task.get('wchan'))

    def test_get_cmd(self):
        """The get() method can return the cmd."""
        self.make_process_file(self.id, 'cmdline', content='cmd')
        self.task.collect_stats()
        self.assertEqual(self.task.get('cmd'), 'cmd')

    def test_equal(self):
        """Two Processes are equal if they have the same pid."""
        other = Process(
            self.id, '{}/{}'.format(self.tempdir.path, self.id))
        different = Process(
            self.id + 1, '{}/{}'.format(self.tempdir.path, self.id))
        self.assertEqual(self.task, other)
        self.assertNotEqual(self.task, different)


class ProcessTests(TestCase):

    def setUp(self):
        super().setUp()
        self.pid = 10
        self.process = Process(
            self.pid, os.path.join(self.tempdir.path, str(self.pid)))

    def test_pid(self):
        """The pid attribute returns the PID."""
        self.assertEqual(self.process.pid, self.pid)

    def test_get_pid(self):
        """The get() method can return the PID."""
        self.process.collect_stats()
        self.assertEqual(self.process.get('pid'), self.pid)

    def test_tasks(self):
        """The list of TIDs for process tasks can be returned."""
        self.make_process_dir(self.pid, 'task/123')
        self.make_process_dir(self.pid, 'task/456')
        self.assertCountEqual(
            self.process.tasks(),
            [Task(123, self, os.path.join(self.tempdir.path, 'task', '123')),
             Task(456, self, os.path.join(self.tempdir.path, 'task', '456'))])


class TaskTests(TestCase):

    def setUp(self):
        super().setUp()
        self.id = 10
        self.process = Process(
            self.id, '{}/{}'.format(self.tempdir.path, self.id))
        self.task = Task(
            self.id, self.process,
            '{}/{}/task/{}'.format(self.tempdir.path, self.id, self.id))

    def test_pid(self):
        """The tid attribute returns the TID."""
        self.assertEqual(self.task.tid, self.id)

    def test_process(self):
        """The parent attribute returns the task parent process."""
        self.assertEqual(self.task.parent, self.process)

    def test_get_pid(self):
        """The get() method can return the PID."""
        self.task.collect_stats()
        self.assertEqual(self.task.get('tid'), self.id)



"""Base classes for reading and parsing text files."""

from collections import Callable

from ..fs import File


class ParsedFile(File):
    """A file whose content is parsed.

    It's intended to be subclassed to provide parsers for specific types of
    files.

    Subclasses must implement the :func:`_parse` method which is called with
    the content of the file and returns the parsed information.

    """

    def parse(self):
        """Read the file and preturn the parsed content."""
        if not self.exists:
            return

        return self._parse(self.read())

    def _parse(self, content):
        """Parse the content of the file.

        .. note::
            Subclasses must implement this method.

        """
        raise NotImplementedError('The parser method must be implemented.')


class SingleLineFile(ParsedFile):
    """A single-line file that can be split into fields.

    The line is split into fields based on a :attr:`separator` (space by
    default).

    If the separator is set to None, the stripped content of the file is
    returned.

    If separator is a callable, it's used to split the field (it's called with
    the line and must return a list of fields).

    Subclasses can define a list of :attr:`fields` and a different
    :attr:`separator`.

    """

    #: The separator to use when splitting the content. If set to :data:`None`,
    #: content is not split.  It can also be set to a `callable` that splits
    #: content.
    separator = ' '

    #: If set, it must be a list, where each element can be
    #:
    #: - a string: the key to use for the value.
    #: - a list of (key, type) tuples:  the value is  converted to the type by
    #:   calling :samp:`type(value)`.
    #: - :data:`None`: the field is ignored.
    fields = None

    def _parse(self, content):
        # Take just fhe first line
        content = content.split('\n')[0]
        if self.separator is None:
            return content

        splitted = self._split(content)
        if self.fields is None:
            return splitted

        fields = self._get_fields()

        # Map fields values to their name converting to the proper type
        return {
            key: field_type(value)
            for (key, field_type), value in zip(fields, splitted)
            if key is not None}

    def _get_fields(self):
        fields = []
        for field in self.fields:
            if field is None:
                field = (None, None)
            elif isinstance(field, str):
                field = (field, str)

            fields.append(field)
        return fields

    def _split(self, content):
        if not content:
            return []

        if isinstance(self.separator, Callable):
            return self.separator(content)

        content = content.strip(self.separator)
        return content.split(self.separator)


class SplittedFile(ParsedFile):
    """A file that is parsed by splitting the content in words.

    It's meant to work with files that have one word per line or a single
    space-separated line.

    Example of content::

      foo
      bar
      baz

    or::

      foo bar baz

    In both cases the result is :samp:`['foo', 'bar', 'baz']`.

    """

    def _parse(self, content):
        lines = content.splitlines()
        if len(lines) == 1:
            return lines[0].split()
        else:
            return lines



"""Base unittest classes."""

import os

from toolrack.testing import TestCase as ToolRackTestCase
from toolrack.testing.fixtures import TempDirFixture


class TestCase(ToolRackTestCase):
    """Base class for tests."""

    def setUp(self):
        super().setUp()
        # A base temporary directory
        self.tempdir = self.useFixture(TempDirFixture())

    def make_process_file(self, pid, name, content='', mode=None):
        """Create a /proc file for the process with the specified content."""
        path = os.path.join(str(pid), name)
        return self.tempdir.mkfile(path=path, content=content, mode=mode)

    def make_process_dir(self, pid, name):
        """Create a subdirectory under a process /proc directory."""
        path = os.path.join(str(pid), name)
        return self.tempdir.mkdir(path=path)



from setuptools import (
    setup,
    find_packages)

from lxstats import (
    __version__,
    __doc__ as description)


config = {
    'name': 'lxstats',
    'version': __version__,
    'license': 'LGPLv3+',
    'description': description,
    'long_description': open('README.rst').read(),
    'author': 'Alberto Donato',
    'author_email': 'alberto.donato@gmail.com',
    'maintainer': 'Alberto Donato',
    'maintainer_email': 'alberto.donato@gmail.com',
    'url': 'https://github.com/albertodonato/lxstats',
    'packages': find_packages(),
    'include_package_data': True,
    "entry_points": {
        "console_scripts": ["procs = lxstats.scripts.procs:script"]},
    'test_suite': 'lxstats',
    'install_requires': ['prettytable', 'toolrack'],
    'keywords': 'linux proc sys performance monitoring system admin',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        ('License :: OSI Approved :: '
         'GNU Lesser General Public License v3 or later (LGPLv3+)'),
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Benchmark',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities']}

setup(**config)



"""Parsers for per-process files under :file:`/proc/[pid]/`."""

import re

from ..text import (
    ParsedFile,
    SingleLineFile)


class ProcPIDCmdline(SingleLineFile):
    """Parse :file:`/proc/[pid]/cmdline`."""

    separator = '\x00'


class ProcPIDStat(SingleLineFile):
    """Parse :file:`/proc/[pid]/stat`, :file:`/proc/[pid]/tasks/[tid]/stat`."""

    fields = (
        ('pid', int),
        ('comm', str),
        ('state', str),
        ('ppid', int),
        ('pgrp', int),
        ('session', int),
        ('tty_nr', int),
        ('tpgid', int),
        ('flags', int),
        ('minflt', int),
        ('cminflt', int),
        ('majflt', int),
        ('cmajflt', int),
        ('utime', int),
        ('stime', int),
        ('cutime', int),
        ('cstime', int),
        ('priority', int),
        ('nice', int),
        ('num_threads', int),
        ('itrealvalue', int),
        ('starttime', int),
        ('vsize', int),
        ('rss', int),
        ('rsslim', int),
        ('startcode', int),
        ('endcode', int),
        ('startstack', int),
        ('kstkesp', int),
        ('kstkeip', int),
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        ('exit_signal', int),
        ('processor', int),
        ('rt_priority', int),
        ('policy', int),
        ('delayacct_blkio_ticks', int),
        ('guest_time', int),
        ('cguest_time', int))

    _re = re.compile('\((.+)\)')

    def separator(self, content):
        """Custom separator to handle spaces in the process commandline."""

        content = content.strip(' ')
        match = self._re.search(content)
        if match is None:
            return content.split()

        # Replace with a version without spaces so split works
        content = self._re.sub('comm', content)
        split = content.split()
        # Replace the original comm value
        split[1] = match.groups()[0]
        return split


class ProcPIDStatm(SingleLineFile):
    """Parse :file:`/proc/[pid]/statm`."""

    fields = (
        ('size', int),
        ('resident', int),
        ('share', int),
        ('text', int),
        ('lib', int),
        ('data', int),
        ('dt', int))


class ProcPIDIo(ParsedFile):
    """Parse :file:`/proc/[pid]/io`."""

    def _parse(self, content):
        # Each line is in the form 'name: count'.
        result = {}
        for line in content.splitlines():
            key, value = line.split(': ', 1)
            result[key] = int(value)
        return result


class ProcPIDSched(ParsedFile):
    """Parse :file:`/proc/[pid]/sched`."""

    _re = re.compile(r'^(\S+)\s+:\s+(\S+)$')

    def _parse(self, content):
        stats = {}
        for line in content.split('\n'):
            match = self._re.match(line)
            if match:
                key, value = match.groups()
                value = float(value) if '.' in value else int(value)
                stats[key] = value

        return stats


class ProcPIDEnviron(ParsedFile):
    """Parse :file:`/proc/[pid]/environ`."""

    def _parse(self, content):
        # Return a dict with the environment.
        environ = {}
        for line in content.split('\x00'):
            if not line:
                continue

            split = line.split('=', 1)
            environ[split[0]] = None if len(split) == 1 else split[1]

        return environ


class ProcPIDCgroup(ParsedFile):
    """Parse :file:`/proc/[pid]/cgroup`."""

    def _parse(self, content):
        result = {}
        for line in content.splitlines():
            hier_id, subsys, control_group = line.split(':')
            result[int(hier_id)] = (subsys.split(','), control_group)

        return result


class ProcPIDStatus(ParsedFile):
    """Parse :file:`/proc/[pid]/status`."""

    _re = re.compile(':\s+')

    def _parse(self, content):
        tokens = [
            self._re.split(line, maxsplit=1) for line in content.splitlines()]
        return {
            key: int(value[:-3]) * 1024 for key, value in tokens
            if value.endswith(' kB')}



"""Access information about running processes."""

from .collection import (
    Collector,
    Collection)
from .process import (
    Process,
    Task)
from .formatter import Formatter
from .filter import CommandLineFilter


__all__ = [
    'Collector', 'Collection', 'Process', 'Task', 'Formatter',
    'CommandLineFilter']



"""Interface to kernel tracing."""

import os

from ..files.sys import TracingDirectory
from .types import TRACER_TYPES


class UnsupportedTracer(Exception):
    """Unsupported tracer type specified."""

    def __init__(self, tracer_type):
        self.type = tracer_type
        super().__init__('Unsupported tracer type: {}'.format(tracer_type))


class Tracing:
    """Interface to kernel tracing."""

    def __init__(self, path='/sys/kernel/debug/tracing'):
        self.path = path

    @property
    def tracers(self):
        """List of current tracing instances in alphabetical order."""
        names = sorted(os.listdir(os.path.join(self.path, 'instances')))
        return [self.get_tracer(name) for name in names]

    def get_tracer(self, name):
        """Return a Tracer.

        If the name doesn't match an existing tracer, a new one is added.

        """
        tracer_path = self._tracer_path(name)
        if not os.path.isdir(tracer_path):
            os.mkdir(tracer_path)
        return Tracer(tracer_path)

    def remove_tracer(self, name):
        """Remove the tracer with the specified name."""
        os.rmdir(self._tracer_path(name))

    def _tracer_path(self, name):
        return os.path.join(self.path, 'instances', name)


class Tracer:
    """A kernel tracing instance."""

    _tracer_types = TRACER_TYPES

    def __init__(self, path):
        self.path = path
        self._dir = TracingDirectory(path)

    def __getattr__(self, attr):
        """Proxy TracerType-specific attributes."""
        return getattr(self._tracer, attr)

    @property
    def name(self):
        """The tracer name."""
        return os.path.basename(self.path)

    @property
    def type(self):
        """'Return the current tracer type."""
        return self._dir['current_tracer'].value

    def set_type(self, tracer_type):
        """Set the type of the tracer."""
        if tracer_type not in self._tracer_types:
            raise UnsupportedTracer(tracer_type)
        self._dir['current_tracer'].set(tracer_type)

    def trace(self):
        """Return content from tracer."""
        with open(os.path.join(self.path, 'trace')) as fd:
            return fd.read()

    def trace_pipe(self):
        """Return an open file descript for the tracing output pipe."""
        return open(os.path.join(self.path, 'trace_pipe'))

    @property
    def enabled(self):
        """Whether the tracer is enabled."""
        return self._dir['tracing_on'].enabled

    def toggle(self, status):
        """Enable or disable the tracer."""
        self._dir['tracing_on'].toggle(status)

    @property
    def options(self):
        """Return a dict with tracing options and their status."""
        return self._dir['trace_options'].options

    def set_option(self, option, value):
        """Set the value of a strcing option."""
        self._dir['trace_options'].toggle(option, value)

    @property
    def _tracer(self):
        """TracerType for the current tracer."""
        return self._tracer_types.get(self.type)
